import flowws
from flowws import Argument as Arg
from geometric_algebra_attention import keras as gala
import numpy as np
import tensorflow as tf
from tensorflow import keras

LAMBDA_ACTIVATIONS = {
    'leakyswish': lambda x: tf.nn.swish(x) - 1e-2*tf.nn.swish(-x),
    'log1pswish': lambda x: tf.math.log1p(tf.nn.swish(x)),
    'sin': tf.sin,
}

NORMALIZATION_LAYERS = {
    None: lambda _: [],
    'none': lambda _: [],
    'batch': lambda _: [keras.layers.BatchNormalization()],
    'layer': lambda _: [keras.layers.LayerNormalization()],
    'momentum': lambda _: [gala.MomentumNormalization()],
    'momentum_layer': lambda _: [gala.MomentumLayerNormalization()],
}

NORMALIZATION_LAYER_DOC = ' (any of {})'.format(
    ','.join(map(str, NORMALIZATION_LAYERS))
)

@flowws.add_stage_arguments
class GalaPDBInverseCoarseGrain(flowws.Stage):
    """Build a geometric attention network for the coarse-grain backmapping task.


    """

    ARGS = [
        Arg('rank', None, int, 2,
            help='Degree of correlations (n-vectors) to consider'),
        Arg('n_dim', '-n', int, 32,
            help='Working dimensionality of point representations'),
        Arg('dilation', None, float, 2,
            help='Working dimension dilation factor for MLP components'),
        Arg('merge_fun', '-m', str, 'concat',
            help='Method to merge point representations'),
        Arg('join_fun', '-j', str, 'concat',
            help='Method to join invariant and point representations'),
        Arg('dropout', '-d', float, 0,
            help='Dropout rate to use, if any'),
        Arg('mlp_layers', None, int, 1,
            help='Number of hidden layers for score/value MLPs'),
        Arg('n_coarse_blocks', '-b', int, 2,
            help='Number of deep blocks to use for coarse-grain '),
        Arg('n_fine_blocks', '-v', int, 2,
            help='Number of deep vector blocks to use'),
        Arg('block_nonlinearity', None, bool, True,
            help='If True, add a nonlinearity to the end of each block'),
        Arg('residual', '-r', bool, True,
            help='If True, use residual connections within blocks'),
        Arg('activation', '-a', str, 'relu',
            help='Activation function to use inside the network'),
        Arg('final_activation', None, str, 'relu',
            help='Final activation function to use within the network'),
        Arg('score_normalization', None, str, 'layer',
            help=('Normalizations to apply to score (attention) function' +
                  NORMALIZATION_LAYER_DOC)),
        Arg('value_normalization', None, str, 'layer',
            help=('Normalizations to apply to value function' +
                  NORMALIZATION_LAYER_DOC)),
        Arg('block_normalization', None, str, 'layer',
            help=('Normalizations to apply to the output of each attention block' +
                  NORMALIZATION_LAYER_DOC)),
        Arg('invariant_value_normalization', None, str, 'momentum',
            help=('Normalizations to apply to value function, before MLP layers' +
                  NORMALIZATION_LAYER_DOC)),
        Arg('equivariant_value_normalization', None, str, 'momentum_layer',
            help=('Normalizations to apply to equivariant results' +
                  NORMALIZATION_LAYER_DOC)),
        Arg('invariant_mode', None, str, 'full',
           help='Attention invariant_mode to use'),
        Arg('covariant_mode', None, str, 'full',
           help='Multivector2MultivectorAttention covariant_mode to use'),
        Arg('use_multivectors', None, bool, False,
            help='If True, use multivector intermediates for calculations'),
        Arg('include_normalized_products', None, bool, False,
           help='Include normalized geometric products in calculations'),
        Arg('convex_covariants', None, bool, False,
            help='Use convex combinations of input points if True'),
        Arg('normalize_equivariant_values', None, bool, False,
           help='If True, multiply vector values by normalized vectors at each attention step'),
    ]

    def run(self, scope, storage):
        use_weights = scope.get('use_bond_weights', False)
        n_dim = self.arguments['n_dim']
        dilation = self.arguments['dilation']
        block_nonlin = self.arguments['block_nonlinearity']
        residual = self.arguments['residual']
        join_fun = self.arguments['join_fun']
        merge_fun = self.arguments['merge_fun']
        dropout = self.arguments['dropout']
        num_blocks = self.arguments['n_coarse_blocks']
        num_vector_blocks = self.arguments['n_fine_blocks']
        rank = self.arguments['rank']
        activation = self.arguments['activation']
        distance_norm = self.arguments.get('normalize_distances', None)
        invar_mode = self.arguments['invariant_mode']
        covar_mode = self.arguments['covariant_mode']
        DropoutLayer = scope.get('dropout_class', keras.layers.Dropout)

        normalization_getter = lambda key: (
            NORMALIZATION_LAYERS[self.arguments.get(key + '_normalization', None)](rank)
        )

        if self.arguments['use_multivectors']:
            Attention = gala.MultivectorAttention
            AttentionVector = gala.Multivector2MultivectorAttention
            AttentionLabeled = gala.LabeledMultivectorAttention
            maybe_upcast_vector = gala.Vector2Multivector()
            maybe_downcast_vector = gala.Multivector2Vector()
        else:
            Attention = gala.VectorAttention
            AttentionVector = gala.Vector2VectorAttention
            AttentionLabeled = gala.LabeledVectorAttention
            maybe_upcast_vector = lambda x: x
            maybe_downcast_vector = lambda x: x

        if self.arguments['activation'] in LAMBDA_ACTIVATIONS:
            activation_layer = lambda: keras.layers.Lambda(
                LAMBDA_ACTIVATIONS[self.arguments['activation']])
        else:
            activation_layer = lambda: keras.layers.Activation(
                self.arguments['activation'])

        if self.arguments['final_activation'] in LAMBDA_ACTIVATIONS:
            final_activation_layer = lambda: keras.layers.Lambda(
                LAMBDA_ACTIVATIONS[self.arguments['final_activation']])
        else:
            final_activation_layer = lambda: keras.layers.Activation(
                self.arguments['final_activation'])

        dilation_dim = int(np.round(n_dim*dilation))

        def make_layer_inputs(x, v):
            nonnorm = (x, v, w_in) if use_weights else (x, v)
            if self.arguments['normalize_equivariant_values']:
                xnorm = keras.layers.LayerNormalization()(x)
                norm = (xnorm, v, w_in) if use_weights else (xnorm, v)
                return [nonnorm] + (rank - 1)*[norm]
            else:
                return rank*[nonnorm]

        def make_scorefun():
            layers = [keras.layers.Dense(dilation_dim)]

            layers.extend(normalization_getter('score'))

            layers.append(activation_layer())
            if dropout:
                layers.append(DropoutLayer(dropout))

            layers.append(keras.layers.Dense(1))
            return keras.models.Sequential(layers)

        def make_valuefun(dim, in_network=True):
            layers = []

            if in_network:
                layers.extend(normalization_getter('invariant_value'))

            layers.append(keras.layers.Dense(dilation_dim))
            layers.extend(normalization_getter('value'))

            layers.append(activation_layer())
            if dropout:
                layers.append(DropoutLayer(dropout))

            layers.append(keras.layers.Dense(dim))
            return keras.models.Sequential(layers)

        def make_block(last_x, last):
            residual_in_x = last_x
            residual_in = last
            if self.arguments['use_multivectors']:
                arg = make_layer_inputs(last_x, last)
                last_x = gala.Multivector2MultivectorAttention(
                    make_scorefun(),
                    make_valuefun(n_dim),
                    make_valuefun(1),
                    False,
                    rank=rank,
                    join_fun=join_fun,
                    merge_fun=merge_fun,
                    invariant_mode=invar_mode,
                    covariant_mode=covar_mode,
                    include_normalized_products=self.arguments['include_normalized_products'],
                    convex_covariants=self.arguments['convex_covariants'],
                )(arg)

            arg = make_layer_inputs(last_x, last)
            last = Attention(
                make_scorefun(),
                make_valuefun(n_dim),
                False,
                rank=rank,
                join_fun=join_fun,
                merge_fun=merge_fun,
                invariant_mode=invar_mode,
                covariant_mode=covar_mode,
                include_normalized_products=self.arguments['include_normalized_products'],
            )(arg)

            if block_nonlin:
                last = make_valuefun(n_dim, in_network=False)(last)

            if residual:
                last = last + residual_in

            for layer in normalization_getter('block'):
                last = layer(last)

            if self.arguments['use_multivectors']:
                if residual:
                    last_x = residual_in_x + last_x
                for layer in normalization_getter('equivariant_value'):
                    last_x = layer(last_x)

            return last_x, last

        def make_vector_block(last_x, last):
            residual_in_x = last_x
            residual_in = last
            arg = [last_x, last]
            last_x = AttentionVector(
                make_scorefun(),
                make_valuefun(n_dim),
                make_valuefun(1),
                False,
                rank=rank,
                join_fun=join_fun,
                merge_fun=merge_fun,
                invariant_mode=invar_mode,
                covariant_mode=covar_mode,
                include_normalized_products=self.arguments['include_normalized_products'],
                convex_covariants=self.arguments['convex_covariants'],
            )(arg)

            arg = [last_x, last]
            last = Attention(
                make_scorefun(),
                make_valuefun(n_dim),
                False,
                rank=rank,
                join_fun=join_fun,
                merge_fun=merge_fun,
                invariant_mode=invar_mode,
                covariant_mode=covar_mode,
                include_normalized_products=self.arguments['include_normalized_products'],
            )(arg)

            if block_nonlin:
                last = make_valuefun(n_dim, in_network=False)(last)

            if residual:
                last = last + residual_in

            for layer in normalization_getter('block'):
                last = layer(last)

            if self.arguments['use_multivectors']:
                if residual:
                    last_x = residual_in_x + last_x
                for layer in normalization_getter('equivariant_value'):
                    last_x = layer(last_x)

            return last_x, last

        x_in = keras.layers.Input((None, 3), name='coarse_rij')
        v_in = keras.layers.Input((None, 2*len(scope['type_names'])), name='coarse_tij')
        label_in = keras.layers.Input((None,), dtype=tf.int32, name='fine_label')
        w_in = None
        inputs = [x_in, v_in, label_in]
        if use_weights:
            w_in = keras.layers.Input((None,), name='wij')
            inputs = [x_in, v_in, w_in, label_in]

        last_x = maybe_upcast_vector(x_in)
        last = keras.layers.Dense(n_dim)(v_in)
        for _ in range(num_blocks):
            last_x, last = make_block(last_x, last)

        atom_embedding = keras.layers.Embedding(
            len(scope['child_type_names']), n_dim, mask_zero=True)(label_in)

        arg = make_layer_inputs(last_x, last)
        arg = [atom_embedding, arg]
        last_x = AttentionLabeled(
            make_scorefun(),
            make_valuefun(n_dim),
            make_valuefun(1),
            True,
            rank=rank,
            join_fun=join_fun,
            merge_fun=merge_fun,
            invariant_mode=invar_mode,
            covariant_mode=covar_mode,
            include_normalized_products=self.arguments['include_normalized_products'],
            convex_covariants=self.arguments['convex_covariants'],
        )(arg)

        last = keras.layers.MultiHeadAttention(8, n_dim)(atom_embedding, last)
        for _ in range(num_vector_blocks):
            pass
            last_x, last = make_vector_block(last_x, last)

        last_x = maybe_downcast_vector(last_x)

        scope['input_symbol'] = inputs
        scope['output'] = last_x
        scope['loss'] = 'mse'



import collections

import numpy as np

try:
    from .keras_metrics import ScaledMAE, ScaledMSE
except ImportError:
    ScaledMAE = ScaledMSE = None

def grab_nested_batch(i, x):
    if isinstance(x, tuple):
        return tuple(grab_nested_batch(i, v) for v in x)
    elif isinstance(x, list):
        return list(grab_nested_batch(i, v) for v in x)
    return x[i]

def nested_batch_len(x):
    if isinstance(x, (tuple, list)):
        return nested_batch_len(x[0])
    return len(x)

def iter_nested_batch(x):
    for i in range(nested_batch_len(x)):
        yield grab_nested_batch(i, x)

class DataMixingPool:
    def __init__(self, size=512, batch_size=16):
        self.size = size
        self.batch_size = batch_size
        self.pending_batches = [[] for _ in range(self.size)]
        self.ready_batches = collections.deque()

    def sample(self, generator, seed=13):
        rng = np.random.default_rng(seed)

        for batch in generator:
            targets = rng.permutation(len(self.pending_batches))
            batch_iter = iter_nested_batch(batch)
            for (target_id, batch_elements) in zip(targets, batch_iter):
                self.pending_batches[target_id].append(batch_elements)
                if len(self.pending_batches[target_id]) >= self.batch_size:
                    self.ready_batches.append(self.pending_batches[target_id])
                    self.pending_batches[target_id] = []

            while self.ready_batches:
                yield self.ready_batches.popleft()




import collections

from .internal import GradientLayer, \
    NeighborhoodReduction, \
    PairwiseValueNormalization, PairwiseVectorDifference, \
    PairwiseVectorDifferenceSum

import flowws
from flowws import Argument as Arg
from geometric_algebra_attention import keras as gala
import numpy as np
import tensorflow as tf
from tensorflow import keras

LAMBDA_ACTIVATIONS = {
    'leakyswish': lambda x: tf.nn.swish(x) - 1e-2*tf.nn.swish(-x),
    'log1pswish': lambda x: tf.math.log1p(tf.nn.swish(x)),
    'sin': tf.sin,
}

NORMALIZATION_LAYERS = {
    None: lambda _, **kwargs: [],
    'none': lambda _, **kwargs: [],
    'batch': lambda _, **kwargs: [keras.layers.BatchNormalization()],
    'layer': lambda _, **kwargs: [keras.layers.LayerNormalization()],
    'momentum': lambda _, **kwargs: [gala.MomentumNormalization(
        momentum=kwargs.get('momentum', .99))],
    'momentum_layer': lambda _, **kwargs: [gala.MomentumLayerNormalization(
        momentum=kwargs.get('momentum', .99))],
}

NORMALIZATION_LAYER_DOC = ' (any of {})'.format(
    ','.join(map(str, NORMALIZATION_LAYERS))
)

@flowws.add_stage_arguments
class GalaMoleculeForceRegression(flowws.Stage):
    """Build a geometric attention network for the molecular force regression task.

    This module specifies the architecture of a network to calculate
    atomic forces given the coordinates and types of atoms in a
    molecule. Conservative forces are computed by calculating the
    gradient of a scalar.

    """

    ARGS = [
        Arg('rank', None, int, 2,
            help='Degree of correlations (n-vectors) to consider'),
        Arg('n_dim', '-n', int, 32,
            help='Working dimensionality of point representations'),
        Arg('dilation', None, float, 2,
            help='Working dimension dilation factor for MLP components'),
        Arg('merge_fun', '-m', str, 'concat',
            help='Method to merge point representations'),
        Arg('join_fun', '-j', str, 'concat',
            help='Method to join invariant and point representations'),
        Arg('dropout', '-d', float, 0,
            help='Dropout rate to use, if any'),
        Arg('mlp_layers', None, int, 1,
            help='Number of hidden layers for score/value MLPs'),
        Arg('n_blocks', '-b', int, 2,
            help='Number of deep blocks to use'),
        Arg('block_nonlinearity', None, bool, True,
            help='If True, add a nonlinearity to the end of each block'),
        Arg('residual', '-r', bool, True,
            help='If True, use residual connections within blocks'),
        Arg('activation', '-a', str, 'swish',
            help='Activation function to use inside the network'),
        Arg('final_activation', None, str, 'swish',
            help='Final activation function to use within the network'),
        Arg('score_normalization', None, str, 'layer',
            help=('Normalizations to apply to score (attention) function' +
                  NORMALIZATION_LAYER_DOC)),
        Arg('value_normalization', None, str, 'layer',
            help=('Normalizations to apply to value function' +
                  NORMALIZATION_LAYER_DOC)),
        Arg('block_normalization', None, str, 'layer',
            help=('Normalizations to apply to the output of each attention block' +
                  NORMALIZATION_LAYER_DOC)),
        Arg('invariant_value_normalization', None, str, 'momentum',
            help=('Normalizations to apply to value function, before MLP layers' +
                  NORMALIZATION_LAYER_DOC)),
        Arg('equivariant_value_normalization', None, str, 'layer',
            help=('Normalizations to apply to equivariant results' +
                  NORMALIZATION_LAYER_DOC)),
        Arg('invariant_mode', None, str, 'single',
           help='Attention invariant_mode to use'),
        Arg('covariant_mode', None, str, 'single',
           help='Multivector2MultivectorAttention covariant_mode to use'),
        Arg('use_multivectors', None, bool, False,
            help='If True, use multivector intermediates for calculations'),
        Arg('include_normalized_products', None, bool, False,
           help='Also include normalized geometric product terms'),
        Arg('normalize_equivariant_values', None, bool, False,
           help='If True, multiply vector values by normalized vectors at each attention step'),
        Arg('center_of_mass', None, bool, False,
            help='If True, use coordinates relative to the molecular center of mass'),
        Arg('reuse_block_layers', None, bool, False,
            help='If True, use the same weights for all block-level attention layers'),
        Arg('normalization_kwargs', None, [(str, eval)], [],
            help='Keyword arguments to pass to normalization functions'),
        Arg('convex_covariants', None, bool, False,
            help='If True, use convex combinations of covariant values'),
        Arg('tied_attention', None, bool, False,
            help='Use tied attention for rotation-equivariant and -invariant signals'),
    ]

    def run(self, scope, storage):
        use_weights = scope.get('use_bond_weights', False)
        n_dim = self.arguments['n_dim']
        dilation = self.arguments['dilation']
        block_nonlin = self.arguments['block_nonlinearity']
        residual = self.arguments['residual']
        join_fun = self.arguments['join_fun']
        merge_fun = self.arguments['merge_fun']
        dropout = self.arguments['dropout']
        num_blocks = self.arguments['n_blocks']
        rank = self.arguments['rank']
        activation = self.arguments['activation']
        distance_norm = self.arguments.get('normalize_distances', None)
        invar_mode = self.arguments['invariant_mode']
        covar_mode = self.arguments['covariant_mode']
        DropoutLayer = scope.get('dropout_class', keras.layers.Dropout)

        normalization_kwargs = dict(self.arguments.get('normalization_kwargs', []))
        normalization_getter = lambda key: (
            NORMALIZATION_LAYERS[self.arguments.get(key + '_normalization', None)](
                rank, **normalization_kwargs)
        )

        if self.arguments['tied_attention']:
            equivariant_layer_builder = lambda: gala.TiedMultivectorAttention(
                make_scorefun(),
                make_valuefun(n_dim),
                make_valuefun(1),
                False,
                rank=rank,
                join_fun=join_fun,
                merge_fun=merge_fun,
                invariant_mode=invar_mode,
                covariant_mode=covar_mode,
                include_normalized_products=self.arguments['include_normalized_products'],
                convex_covariants=self.arguments['convex_covariants'],
            )
        else:
            equivariant_layer_builder = lambda: gala.Multivector2MultivectorAttention(
                make_scorefun(),
                make_valuefun(n_dim),
                make_valuefun(1),
                False,
                rank=rank,
                join_fun=join_fun,
                merge_fun=merge_fun,
                invariant_mode=invar_mode,
                covariant_mode=covar_mode,
                include_normalized_products=self.arguments['include_normalized_products'],
                convex_covariants=self.arguments['convex_covariants'],
            )
        invariant_layer_builder = lambda: Attention(
            make_scorefun(),
            make_valuefun(n_dim),
            False,
            rank=rank,
            join_fun=join_fun,
            merge_fun=merge_fun,
            invariant_mode=invar_mode,
            covariant_mode=covar_mode,
            include_normalized_products=self.arguments['include_normalized_products'],
        )

        if self.arguments['use_multivectors']:
            Attention = gala.MultivectorAttention
            AttentionVector = gala.Multivector2MultivectorAttention
            maybe_upcast_vector = gala.Vector2Multivector()
            maybe_downcast_vector = gala.Multivector2Vector()
        else:
            Attention = gala.VectorAttention
            AttentionVector = gala.Vector2VectorAttention
            maybe_upcast_vector = lambda x: x


        if self.arguments['activation'] in LAMBDA_ACTIVATIONS:
            activation_layer = lambda: keras.layers.Lambda(
                LAMBDA_ACTIVATIONS[self.arguments['activation']])
        else:
            activation_layer = lambda: keras.layers.Activation(
                self.arguments['activation'])

        if self.arguments['final_activation'] in LAMBDA_ACTIVATIONS:
            final_activation_layer = lambda: keras.layers.Lambda(
                LAMBDA_ACTIVATIONS[self.arguments['final_activation']])
        else:
            final_activation_layer = lambda: keras.layers.Activation(
                self.arguments['final_activation'])

        dilation_dim = int(np.round(n_dim*dilation))

        def make_layer_inputs(x, v):
            nonnorm = (x, v, w_in) if use_weights else (x, v)
            if self.arguments['normalize_equivariant_values']:
                xnorm = keras.layers.LayerNormalization()(x)
                norm = (xnorm, v, w_in) if use_weights else (xnorm, v)
                return [nonnorm] + (rank - 1)*[norm]
            else:
                return rank*[nonnorm]

        def make_scorefun():
            layers = [keras.layers.Dense(dilation_dim)]

            layers.extend(normalization_getter('score'))

            layers.append(activation_layer())
            if dropout:
                layers.append(DropoutLayer(dropout))

            layers.append(keras.layers.Dense(1))
            return keras.models.Sequential(layers)

        def make_valuefun(dim, in_network=True):
            layers = []

            if in_network:
                layers.extend(normalization_getter('invariant_value'))

            layers.append(keras.layers.Dense(dilation_dim))
            layers.extend(normalization_getter('value'))

            layers.append(activation_layer())
            if dropout:
                layers.append(DropoutLayer(dropout))

            layers.append(keras.layers.Dense(dim))
            return keras.models.Sequential(layers)

        if self.arguments['reuse_block_layers']:
            the_equivariant_layer = equivariant_layer_builder()
            equivariant_getter = collections.defaultdict(lambda: the_equivariant_layer)
            the_invariant_layer = invariant_layer_builder()
            invariant_getter = collections.defaultdict(lambda: the_invariant_layer)
        else:
            equivariant_getter = collections.defaultdict(equivariant_layer_builder)
            invariant_getter = collections.defaultdict(invariant_layer_builder)

        def make_block(i, last_x, last):
            residual_in_x = last_x
            residual_in = last
            if self.arguments['tied_attention']:
                arg = make_layer_inputs(last_x, last)
                (last_x, last) = equivariant_getter[i](arg)
            else:
                if self.arguments['use_multivectors']:
                    arg = make_layer_inputs(last_x, last)
                    last_x = equivariant_getter[i](arg)

                arg = make_layer_inputs(last_x, last)
                last = invariant_getter[i](arg)

            if block_nonlin:
                last = make_valuefun(n_dim, in_network=False)(last)

            if residual:
                last = last + residual_in

            for layer in normalization_getter('block'):
                last = layer(last)

            if self.arguments['use_multivectors']:
                if residual:
                    last_x = residual_in_x + last_x
                for layer in normalization_getter('equivariant_value'):
                    last_x = layer(last_x)

            return last_x, last

        x_in = keras.layers.Input((scope['neighborhood_size'], 3))
        v_in = keras.layers.Input((scope['neighborhood_size'], scope['num_types']))
        w_in = None
        inputs = [x_in, v_in]
        if use_weights:
            w_in = keras.layers.Input((None,), name='wij')
            inputs = [x_in, v_in, w_in]

        if self.arguments['center_of_mass']:
            mask = tf.math.reduce_any(v_in != 0, axis=-1)
            sum_part = NeighborhoodReduction()(x_in, mask=mask)
            count = tf.math.reduce_sum(tf.cast(mask, 'float32'), axis=-1, keepdims=True)
            delta_x = x_in - (sum_part/count[..., None])[..., None, :]
            delta_v = v_in
        else:
            delta_x = PairwiseVectorDifference()(x_in)
            delta_v = PairwiseVectorDifferenceSum()(v_in)

        last_x = maybe_upcast_vector(delta_x)
        last = keras.layers.Dense(n_dim)(delta_v)
        for i in range(num_blocks):
            last_x, last = make_block(i, last_x, last)

        arg = make_layer_inputs(last_x, last)
        (last, ivs, att) = Attention(
            make_scorefun(),
            make_valuefun(n_dim),
            True,
            name='final_attention',
            rank=rank,
            join_fun=join_fun,
            merge_fun=merge_fun,
            invariant_mode=invar_mode,
            covariant_mode=covar_mode,
            include_normalized_products=self.arguments['include_normalized_products'],
        )(arg, return_invariants=True, return_attention=True)

        last = keras.layers.Dense(dilation_dim, name='final_mlp')(last)
        last = final_activation_layer()(last)
        if not self.arguments['center_of_mass']:
            last = NeighborhoodReduction()(last)
        last = keras.layers.Dense(1, name='energy_projection', use_bias=False)(last)
        last = GradientLayer()((last, x_in))

        scope['input_symbol'] = [x_in, v_in]
        scope['output'] = last
        scope['loss'] = 'mse'
        scope['attention_model'] = keras.models.Model([x_in, v_in], att)
        scope['invariant_model'] = keras.models.Model([x_in, v_in], ivs)



import collections
import contextlib
import json

from emmet.core.summary import HasProps
import flowws
from flowws import Argument as Arg
import gtar
from mp_api.client import MPRester
import numpy as np

@flowws.add_stage_arguments
class ImportMatProjChargeDensity(flowws.Stage):
    """Import charge density data from the Materials Project.

    This module downloads and processes data from the Materials
    Project charge density API
    (https://materialsproject.org/ml/charge_densities). Data are
    imported, adapted into a form suitable for analysis with freud,
    and saved into a getar file.

    """

    ARGS = [
        Arg('api_key', None, str,
            help='API key to access materials project database'),
        Arg('output_file', '-o', str, 'materials_project_charge.sqlite',
            help='Name of output getar file to produce'),
        Arg('chunk_size', '-b', int, 1000,
            help='Chunk size for API requests'),
        Arg('num_chunks', '-n', int,
            help='Number of chunks of data to request'),
        Arg('elements', '-e', [str],
            help='Element symbols to filter search with'),
        Arg('query_kwargs', '-q', [(str, eval)],
            help='Additional query keyword arguments'),
    ]

    def run(self, scope, storage):
        api_key = self.arguments['api_key']
        num_chunks = self.arguments.get('num_chunks', None)
        all_types = set()
        density_stds = []

        with contextlib.ExitStack() as stack:
            mpr = stack.enter_context(MPRester(api_key=api_key))

            kwargs = {}
            if self.arguments.get('elements', None):
                kwargs['elements'] = self.arguments['elements']

            for (key, val) in dict(self.arguments.get('query_kwargs', [])).items():
                kwargs[key] = val

            docs = mpr.summary.search(
                num_chunks=num_chunks,
                chunk_size=self.arguments['chunk_size'],
                has_props=[HasProps.charge_density],
                **kwargs
            )

            traj = stack.enter_context(gtar.GTAR(self.arguments['output_file'], 'w'))

            for doc in docs:
                density = mpr.get_charge_density_from_material_id(doc.material_id)
                data_dict = self.write_gtar_entry(traj, doc, density)
                all_types.update(data_dict['type_names'])
                density_stds.append(data_dict['std'])

            traj.writePath('all_type_names.json', json.dumps(sorted(list(all_types))))
            traj.writePath('density_scale.f32.uni', np.mean(density_stds))

    @staticmethod
    def write_gtar_entry(traj, doc, density):
        group = str(doc.material_id)

        structure = density.structure
        box_matrix = structure.lattice.matrix.T

        Nx = len(density.xpoints)
        Ny = len(density.ypoints)
        Nz = len(density.zpoints)

        assert np.allclose(np.linspace(0, 1, Nx), density.xpoints)
        assert np.allclose(np.linspace(0, 1, Ny), density.ypoints)
        assert np.allclose(np.linspace(0, 1, Nz), density.zpoints)

        grid = (Nx, Ny, Nz)

        positions = structure.cart_coords
        all_type_names = [t.name for t in structure.species]
        type_name_map = collections.defaultdict(lambda: len(type_name_map))
        types = [type_name_map[name] for name in all_type_names]
        type_names = [t for (t, i) in
                      sorted(type_name_map.items(), key=lambda x: x[::-1])]

        density_values = density.data['total']
        std = np.std(density_values)

        traj.writePath('{}/grid.u32.uni'.format(group), grid)
        traj.writePath('{}/box_matrix.f32.uni'.format(group), box_matrix)
        traj.writePath('{}/point_density.f32.uni'.format(group), density_values)
        traj.writePath('{}/position.f32.ind'.format(group), positions)
        traj.writePath('{}/type.u32.ind'.format(group), types)
        traj.writePath('{}/type_names.json'.format(group), json.dumps(type_names))
        return dict(type_names=type_names, std=std)



import collections
import json

import flowws
from flowws import Argument as Arg
import freud
import gtar
import numpy as np
import pyriodic
import rowan

from .internal import DataMixingPool

ChargeDensityRecord = collections.namedtuple('ChargeDensityRecord',
    ['grid', 'box_matrix', 'point_density', 'positions', 'types', 'type_names'])

def _fix_box(box_matrix, dtype=np.float32):
    """Convert box into a right-handed coordinate frame with
    only upper triangular entries. Return a new box, rotation
    quaternion, and transformation function for coordinates."""
    # First use QR decomposition to compute the new basis
    Q, R = np.linalg.qr(box_matrix)
    Q = Q.astype(dtype)
    R = R.astype(dtype)

    if (not np.allclose(Q, np.eye(3)) or np.any(np.diag(R) < 0)):
        # Since we'll be performing a quaternion operation,
        # we have to ensure that Q is a pure rotation
        sign = np.linalg.det(Q)
        Q = Q*sign
        R = R*sign

        quat = rowan.from_matrix(Q.T)

        # Now we have to ensure that the box is right-handed. We
        # do this as a second step to avoid introducing reflections
        # into the rotation matrix before making the quaternion
        signs = np.diag(np.diag(np.where(R < 0, -np.ones(R.shape), np.ones(R.shape))))
        box = R.dot(signs)
    else:
        quat = np.asarray([1., 0, 0, 0])
        signs = np.eye(3)
        box = box_matrix

    # Construct the box
    Lx, Ly, Lz = np.diag(box).flatten().tolist()
    xy = box[0, 1]/Ly
    xz = box[0, 2]/Lz
    yz = box[1, 2]/Lz
    box = freud.box.Box(Lx=Lx, Ly=Ly, Lz=Lz, xy=xy, xz=xz, yz=yz)

    transform = lambda r, Q=Q, signs=signs: r.dot(Q).dot(signs)
    return box, quat, transform

@flowws.add_stage_arguments
class MatProjChargeDensity(flowws.Stage):
    """Read charge density data from a Materials Project archive.

    This module loads a previously-saved getar archive of materials
    project data (using :py:class:`ImportMatProjChargeDensity`) and
    generates point clouds for charge density regression.
    """

    ARGS = [
        Arg('filename', '-f', str, 'materials_project_charge.sqlite',
            help='Filename of imported data to load'),
        Arg('batch_size', '-b', int, 16,
            help='Number of point clouds in each batch'),
        Arg('pool_size', '-p', int, 512,
            help='Number of batches to mix in memory pools'),
        Arg('num_neighbors', '-n', int, 16,
            help='Number of atoms to use for charge density calculation'),
        Arg('seed', '-s', int, 13,
            help='RNG seed for dataset generation'),
        Arg('validation_split', '-v', float, .3,
            help='Fraction of records to use for validation'),
        Arg('samples_per_system', None, int, 1024,
            help='Number of points to draw for each system'),
        Arg('no_keras', '-k', bool, False,
            help='If True, don\'t load/import keras'),
    ]

    def run(self, scope, storage):
        freud.parallel.set_num_threads(1)
        rng = np.random.default_rng(self.arguments['seed'])

        traj = gtar.GTAR(self.arguments['filename'], 'r')

        all_type_names = traj.readStr('all_type_names.json')
        all_type_names = json.loads(all_type_names)
        type_name_map = {t: i + 1 for (i, t) in enumerate(all_type_names)}

        density_scale = traj.readPath('density_scale.f32.uni')

        all_groups = list(sorted(
            {rec.getGroup() for rec in traj.getRecordTypes() if rec.getGroup()}))
        group_indices = rng.permutation(len(all_groups))

        N_val = int(self.arguments['validation_split']*len(group_indices))
        train_groups = [all_groups[i] for i in group_indices[N_val:]]
        val_groups = [all_groups[i] for i in group_indices[:N_val]]

        train_pool = DataMixingPool(self.arguments['pool_size'], self.arguments['batch_size'])
        train_gen_ = self.dataset_generator(
            traj, train_groups, rng.integers(0, 2**32), type_name_map,
            self.arguments['samples_per_system'], self.arguments['num_neighbors'],
            density_scale)
        train_gen = map(
            self.collate_batch, train_pool.sample(train_gen_, rng.integers(0, 2**32)))
        val_pool = DataMixingPool(self.arguments['pool_size'], self.arguments['batch_size'])
        val_gen_ = self.dataset_generator(
            traj, val_groups, rng.integers(0, 2**32), type_name_map,
            self.arguments['samples_per_system'], self.arguments['num_neighbors'],
            density_scale)
        val_gen = map(
            self.collate_batch, val_pool.sample(val_gen_, rng.integers(0, 2**32)))

        if not self.arguments['no_keras']:
            from .internal import ScaledMSE, ScaledMAE
            metrics = scope.setdefault('metrics', [])
            scaled_mae = ScaledMAE(density_scale, name='scaled_mae')
            metrics.append(scaled_mae)

        scope['loss'] = 'mse'
        scope['train_generator'] = train_gen
        scope['validation_generator'] = val_gen
        scope['type_map'] = dict(type_name_map)
        scope['max_types'] = len(type_name_map) + 1
        scope['type_embedding_size'] = scope['max_types']

    @staticmethod
    def collate_batch(batch):
        rijs, tjs, ys = [], [], []
        for (rij, tj, y) in batch:
            rijs.append(rij)
            tjs.append(tj)
            ys.append(y)

        return (np.array(rijs), np.array(tjs)), np.array(ys)

    @classmethod
    def dataset_generator(cls, traj, groups, seed, type_name_map,
                          samples_per_system, num_neighbors, density_scale):
        rng = np.random.default_rng(seed)
        query_args = dict(
            mode='nearest', num_neighbors=num_neighbors, exclude_ii=True)

        while True:
            rng.shuffle(groups)
            for group in groups:
                record = cls.read_gtar_entry(traj, group)
                (box, coords, positions) = cls.translate_to_freud(
                    record.grid, record.box_matrix, record.positions)
                boxarr = [box.Lx, box.Ly, box.Lz, box.xy, box.xz, box.yz]
                struc = pyriodic.Structure(positions, record.types, boxarr)
                Nrep = 64//2
                nlist = []

                while len(nlist) < samples_per_system*num_neighbors:
                    Nrep *= 2
                    struc = struc.replicate_upto(Nrep)

                    coord_idx = rng.permutation(len(record.point_density))
                    coord_idx = coord_idx[:samples_per_system]
                    selected_density = record.point_density[coord_idx]/density_scale
                    selected_coords = coords[coord_idx]

                    aabb = freud.locality.AABBQuery(struc.box, struc.positions)
                    nq = aabb.query(selected_coords, query_args)
                    nlist = nq.toNeighborList()

                ri = selected_coords[nlist.query_point_indices]
                rj = struc.positions[nlist.point_indices]
                rij = box.wrap(rj - ri)

                type_names = record.type_names
                tj = [type_name_map.get(type_names[t], 0)
                      for t in struc.types[nlist.point_indices]]
                tj = np.eye(len(type_name_map) + 1)[tj]

                rij = rij.reshape((samples_per_system, num_neighbors, 3))
                tj = tj.reshape((samples_per_system, num_neighbors, -1))
                selected_density = selected_density.reshape(
                    (samples_per_system, 1))

                yield rij, tj, selected_density

    @staticmethod
    def read_gtar_entry(traj, group):
        grid = traj.readPath('{}/grid.u32.uni'.format(group))
        box_matrix = traj.readPath('{}/box_matrix.f32.uni'.format(group)).reshape((3, 3))
        point_density = traj.readPath('{}/point_density.f32.uni'.format(group))
        positions = traj.readPath('{}/position.f32.ind'.format(group))
        types = traj.readPath('{}/type.u32.ind'.format(group))
        type_names = traj.readPath('{}/type_names.json'.format(group))

        return ChargeDensityRecord(
            grid, box_matrix, point_density, positions, types, type_names)

    @staticmethod
    def translate_to_freud(grid, box_matrix, positions):
        (box, quat, fix_transform) = _fix_box(box_matrix)

        grid_x, grid_y, grid_z = [np.linspace(0, 1, N) for N in grid]
        frac_coords = np.concatenate(
            [v[..., None] for v in np.meshgrid(
                grid_x, grid_y, grid_z)], axis=-1)
        frac_coords = np.transpose(frac_coords, (1, 0, 2, 3)).reshape((-1, 3))
        coords = box.wrap(fix_transform(np.dot(frac_coords, box_matrix.T)))

        positions = box.wrap(fix_transform(positions))

        return (box, coords, positions)



from flowws import try_to_import

ImportMatProjChargeDensity = try_to_import(
    '.ImportMatProjChargeDensity', 'ImportMatProjChargeDensity', __name__)
ImportOCP = try_to_import(
    '.ImportOCP', 'ImportOCP', __name__)
MatProjChargeDensity = try_to_import(
    '.MatProjChargeDensity', 'MatProjChargeDensity', __name__)
from .MD17 import MD17
ModelNet = try_to_import('.ModelNet', 'ModelNet', __name__)
from .PDBCache import PDBCache
PDBCoarseGrained = try_to_import('.PDBCoarseGrained', 'PDBCoarseGrained', __name__)
PyriodicDataset = try_to_import('.PyriodicDataset', 'PyriodicDataset', __name__)
from .RMD17 import RMD17



from flowws import try_to_import

CrystalStructureClassification = try_to_import(
    '.CrystalStructureClassification', 'CrystalStructureClassification', __name__)
GalaMoleculeForceRegression = try_to_import(
    '.GalaMoleculeForceRegression', 'GalaMoleculeForceRegression', __name__)
GalaPDBInverseCoarseGrain = try_to_import(
    '.GalaPDBInverseCoarseGrain', 'GalaPDBInverseCoarseGrain', __name__)
MoleculeForceRegression = try_to_import(
    '.MoleculeForceRegression', 'MoleculeForceRegression', __name__)
PDBInverseCoarseGrain = try_to_import(
    '.PDBInverseCoarseGrain', 'PDBInverseCoarseGrain', __name__)
PDBInverseCoarseGrainTransformer = try_to_import(
    '.PDBInverseCoarseGrainTransformer', 'PDBInverseCoarseGrainTransformer', __name__)



import collections
import json
import pickle

import flowws
from flowws import Argument as Arg
import gtar
import lmdb
import mendeleev
import numpy as np

from .MatProjChargeDensity import _fix_box

try:
    from tqdm import tqdm
except ImportError:
    def tqdm(xs):
        for x in xs:
            yield x

FixedFrame = collections.namedtuple(
    'FixedFrame', ['cell', 'pos', 'atomic_numbers', 'tags', 'force', 'y', 'y_relaxed'])

def fix_frame(frame):
    try:
        _ = frame.pos
        return frame
    except RuntimeError: # mismatched pyg version
        framedict = vars(frame)
        args = [framedict.get(key, None) for key in FixedFrame._fields]
        return FixedFrame(*args)

@flowws.add_stage_arguments
class ImportOCP(flowws.Stage):
    ARGS = [
        Arg('lmdb_file', '-l', str,
            help='Path for data LMDB file to load'),
        Arg('gtar_file', '-g', str, 'ocp.sqlite',
            help='Output getar filename to save'),
        Arg('energy_fit_rows', None, int, 64000,
            help='Maximum number of type histograms to fit energy with'),
        Arg('name_adsorbates', None, bool, True,
            help='If True, give adsorbate atoms the "_ads" suffix'),
    ]

    def run(self, scope, storage):
        gtar_path = self.arguments['gtar_file']
        lmdb_path = self.arguments['lmdb_file']
        env = lmdb.open(lmdb_path, readonly=True, subdir=False)

        fit_rows = self.arguments['energy_fit_rows']
        name_adsorbates = self.arguments['name_adsorbates']

        type_names = ['NONE'] + [e.symbol for e in mendeleev.get_all_elements()]
        type_map = collections.defaultdict(lambda: len(type_map))
        _ = [type_map[t] for t in type_names]

        with env.begin() as txn, gtar.GTAR(gtar_path, 'w') as traj:
            try:
                N = pickle.loads(txn.get('length'.encode()))
            except TypeError: # length not saved
                N = env.stat()['entries']

            rows = {}

            for i in tqdm(range(N)):
                frame = pickle.loads(txn.get(str(i).encode()))
                frame = fix_frame(frame)

                try:
                    energy = frame.y
                    if energy is None:
                        energy = frame.y_relaxed
                except AttributeError:
                    energy = frame.y_relaxed

                (fbox, quat, transform) = _fix_box(frame.cell[0].T)
                boxarr = [fbox.Lx, fbox.Ly, fbox.Lz, fbox.xy, fbox.xz, fbox.yz]

                types = np.asarray(frame.atomic_numbers, dtype=np.int32)
                if name_adsorbates:
                    for j in np.where(frame.tags == 2)[0]:
                        types[j] = type_map['{}_ads'.format(type_names[types[j]])]
                traj.writePath('frames/{}/position.f32.ind'.format(i),
                               transform(np.array(frame.pos)))
                traj.writePath('frames/{}/force.f32.ind'.format(i),
                               transform(np.array(frame.force)))
                traj.writePath('frames/{}/type.u32.ind'.format(i), types)
                traj.writePath('frames/{}/energy.f32.uni'.format(i), energy)
                traj.writePath('frames/{}/box.f32.uni'.format(i), boxarr)
                traj.writePath('frames/{}/num_adsorbate_atoms.u32.uni'.format(i),
                               sum(frame.tags == 2))

                fit_x = np.bincount(types)
                rows[np.random.default_rng(i).integers(fit_rows)] = (
                    fit_x, energy)

            fit_x = np.zeros((len(rows), len(type_map) + 1))
            fit_y = np.zeros(len(fit_x))
            for i, (_, (x, y)) in enumerate(sorted(rows.items())):
                fit_x[i, :len(x)] = x
                fit_y[i] = y
            fit_x[:, -1] = 1
            (x, res, rank, s) = np.linalg.lstsq(fit_x, fit_y)
            x[np.sum(fit_x, axis=0) == 0] = 0
            traj.writePath('linear_energy_fit.f32.uni', x)

            inv_type_name_map = {i: n for (n, i) in type_map.items()}
            type_names = [inv_type_name_map[i] for i in range(len(inv_type_name_map))]
            traj.writeStr('type_names.json', json.dumps(type_names))



import functools

import flowws
from flowws import Argument as Arg
from flowws.Argument import Range
import gtar
import numpy as np

from ..data.MatProjChargeDensity import MatProjChargeDensity

@flowws.add_stage_arguments
class MatProjChargeDensityViewer(flowws.Stage):
    """Visualize charge density from an imported Materials Project file.

    """

    ARGS = [
        Arg('filename', '-f', str, 'materials_project_charge.sqlite',
            help='Filename of imported data to load'),
        Arg('point', '-p', int, -1, valid_values=Range(-1, 16000, True),
            help='Select an individual point by index, or view the whole system if negative'),
        Arg('filter_points', '-n', int, 1024, valid_values=Range(0, 16000, True),
            help='Number of high-density points to view if given'),
        Arg('color_scale', None, float, .25, valid_values=flowws.Range(0, 10, True),
            help='Factor to scale point color RGB intensities by'),
        Arg('diameter', '-d', float, .1,
            help='Point sphere diameter'),
        Arg('group', '-g', str,
            help='Group to visualize'),
    ]

    def run(self, scope, storage):
        if getattr(self, 'gtar_file', None) is None:
            self.gtar_file = gtar.GTAR(self.arguments['filename'], 'r')
            groups = {rec.getGroup() for rec in self.gtar_file.getRecordTypes()
                      if rec.getGroup()}
            groups = list(sorted(groups))
            self.arg_specifications['group'].valid_values = groups
            self.arguments.setdefault('group', groups[0])

        self.record = MatProjChargeDensity.read_gtar_entry(
            self.gtar_file, self.arguments['group'])
        self.freud_system = (box, _, positions) = MatProjChargeDensity.translate_to_freud(
            self.record.grid, self.record.box_matrix, self.record.positions)

        scope['position'] = positions
        scope['type'] = self.record.types
        scope['box'] = [box.Lx, box.Ly, box.Lz, box.xy, box.xz, box.yz]
        scope.setdefault('visuals', []).append(self)

    def draw_plato(self):
        import plato, plato.draw as draw

        record = self.record
        (box, coords, positions) = self.freud_system

        point_thetas = record.point_density.copy()
        point_thetas -= np.min(point_thetas)
        point_thetas /= np.max(point_thetas)
        point_thetas = .8*point_thetas + .2*(1 - point_thetas)

        if self.arguments['filter_points']:
            sortidx = np.argsort(-record.point_density)
            filt = sortidx[:self.arguments['filter_points']]
            coords = coords[filt]
            point_thetas = point_thetas[filt]

        point_colors = plato.cmap.cubehelix(point_thetas)
        point_colors[:, :3] *= self.arguments['color_scale']

        point_prim = draw.Spheres(
            positions=coords, colors=point_colors, diameters=self.arguments['diameter'])

        atom_colors = np.ones((len(positions), 4))
        atom_colors[:, :3] = plato.cmap.cubeellipse_intensity(
            record.types.astype(np.float32))
        atom_prim = draw.Spheres(positions=positions, colors=atom_colors)

        box_prim = draw.Box.from_box(box, color=(.5, .5, .5, 1), width=.1)

        features = dict(additive_rendering=True)
        return draw.Scene([point_prim, atom_prim, box_prim], features=features)



import itertools
import os
import random
import re
import subprocess

import flowws
from flowws import Argument as Arg
import gtar
import numpy as np

header_pattern = re.compile(r'^OFF\s*(?P<vertex>\d+)\s+(?P<mesh>\d+)\s+\d+\s+')

class GTARDataset:
    def __init__(self, fname, neighborhood_size, validation_fraction=0, seed=13):
        self.handle = gtar.GTAR(fname, 'r')
        self.neighborhood_size = neighborhood_size
        self.validation_fraction = validation_fraction
        self.seed = seed

        group_indices = {}
        group_records = {}
        for rec in self.handle.getRecordTypes():
            if rec.getName() != 'vertex':
                continue
            frames = self.handle.queryFrames(rec)
            group_indices[rec.getGroup()] = frames
            group_records[rec.getGroup()] = rec
        self.group_indices = group_indices
        self.group_records = group_records

        if validation_fraction:
            rng = random.Random(seed)
            for g in list(self.group_indices):
                if 'train' not in g:
                    continue

                index_values = self.group_indices[g]
                indices = list(range(len(index_values)))
                rng.shuffle(indices)
                N = max(1, int((1 - validation_fraction)*len(indices)))

                train_indices = [index_values[i] for i in sorted(indices[:N])]
                val_indices = [index_values[i] for i in sorted(indices[N:])]

                val_g = g.replace('train', 'val')
                if val_indices:
                    self.group_indices[val_g] = val_indices
                self.group_indices[g] = train_indices
                self.group_records[val_g] = self.group_records[g]

        labels = set()
        for group in group_indices:
            label = group.split('/')[0]
            labels.add(label)
        self.labels = list(sorted(labels))
        self.label_map = {l: i for (i, l) in enumerate(self.labels)}
        self.num_labels = len(self.labels)

        self.train_groups = [g for g in sorted(self.group_indices) if '/train' in g]
        self.val_groups = [g for g in sorted(self.group_indices) if '/val' in g]
        self.test_groups = [g for g in sorted(self.group_indices) if '/test' in g]

    def _generator(self, dataset, batch_size, seed=13):
        rng = np.random.default_rng(seed)
        groups = getattr(self, '{}_groups'.format(dataset))

        while True:
            rs = np.zeros((batch_size, self.neighborhood_size, 3))
            vs = np.zeros((batch_size, self.neighborhood_size, 1))
            ys = np.zeros(batch_size, dtype=np.int32)

            for i in range(batch_size):
                y = rng.choice(len(groups))
                group = groups[y]
                indices = self.group_indices[group]
                index = indices[rng.choice(len(indices))]

                vertices = self.handle.getRecord(self.group_records[group], index)
                vertices = vertices.reshape((-1, 3))

                vertices -= np.mean(vertices, axis=0, keepdims=True)
                scale = np.mean(np.linalg.norm(vertices, axis=-1))

                vertex_choice = rng.choice(
                    len(vertices), self.neighborhood_size,
                    replace=(self.neighborhood_size > len(vertices)))
                rs[i, :len(vertex_choice)] = vertices[vertex_choice]/scale
                vs[i, :len(vertex_choice)] = 1
                ys[i] = y

            yield ((rs, vs), ys)

    def train_generator(self, batch_size, seed=13):
        yield from self._generator('train', batch_size, seed)

    def test_generator(self, batch_size, seed=13):
        yield from self._generator('test', batch_size, seed)

    def val_generator(self, batch_size, seed=13):
        yield from self._generator('val', batch_size, seed)

@flowws.add_stage_arguments
class ModelNet(flowws.Stage):
    """Load the ModelNet40 dataset."""

    ARGS = [
        Arg('cache_dir', '-c', str, '/tmp',
            help='Directory to store dataset'),
        Arg('neighborhood_size', '-n', int, 12,
            help='Number of points to sample for each cloud'),
        Arg('seed', '-s', int, 13,
            help='Random seed for dataset generation'),
        Arg('validation_fraction', '-v', float, 0,
            help='Fraction of training data to use for validation set'),
        Arg('batch_size', '-b', int, 8,
            help='Size of point cloud batches'),
    ]

    def run(self, scope, storage):
        dataset_fname = self._get_dataset(self.arguments['cache_dir'])
        dataset = GTARDataset(
            dataset_fname, self.arguments['neighborhood_size'],
            self.arguments['validation_fraction'], self.arguments['seed'])

        scope['type_names'] = dataset.labels
        scope['type_name_map'] = dataset.label_map
        scope['train_generator'] = dataset.train_generator(
            self.arguments['batch_size'], self.arguments['seed'] + 1)
        scope['test_generator'] = dataset.test_generator(
            self.arguments['batch_size'], self.arguments['seed'] + 2)
        if self.arguments['validation_fraction']:
            scope['validation_generator'] = dataset.val_generator(
                self.arguments['batch_size'], self.arguments['seed'] + 3)
        scope['neighborhood_size'] = self.arguments['neighborhood_size']
        scope['num_classes'] = len(dataset.labels)

    @classmethod
    def _get_dataset(cls, cache_dir):
        fname = os.path.join(cache_dir, 'modelnet40_gtar.zip')

        if os.path.exists(fname):
            return fname

        url = 'https://modelnet.cs.princeton.edu/ModelNet40.zip'
        cmd = ['wget', '-c', '-P', cache_dir, url]
        subprocess.check_call(cmd)

        source_fname = os.path.join(cache_dir, 'ModelNet40.zip')

        with gtar.GTAR(source_fname, 'r') as src, gtar.GTAR(fname, 'w') as dst:
            for rec in src.getRecordTypes():
                if 'off' not in rec.getName():
                    continue
                for index in src.queryFrames(rec):
                    rec.setIndex(index)
                    cls._translate_record(src, dst, rec)
            pass

        os.remove(source_fname)
        return fname

    @staticmethod
    def _translate_record(src, dst, rec):
        new_group = '/'.join(rec.getGroup().split('/')[1:])
        new_index = rec.getName().split('_')[1].split('.')[0]

        contents = src.getRecord(rec)
        match = header_pattern.match(contents)
        if not match:
            raise ValueError(contents[:32])

        contents = contents[match.end():]
        lines = contents.splitlines()

        # skip "OFF" text
        vertex_count = int(match.group('vertex'))
        mesh_count = int(match.group('mesh'))
        vertices = [list(map(float, line.split())) for line in lines[:vertex_count]]
        mesh_lines = [list(map(int, line.split())) for line in lines[vertex_count:]]
        mesh_stream = list(itertools.chain.from_iterable(mesh_lines))

        assert vertex_count + mesh_count == len(lines)

        new_path = '/'.join([new_group, 'frames', new_index, 'vertex.f32.ind'])
        dst.writePath(new_path, vertices)

        new_path = '/'.join([new_group, 'frames', new_index, 'mesh.u32.uni'])
        dst.writePath(new_path, mesh_stream)



from flowws import try_to_import

algebra = try_to_import('.', 'algebra', __name__)
from . import data
layers = try_to_import('.', 'layers', __name__)
models = try_to_import('.', 'models', __name__)
from . import viz
